#ifndef _FILE_OFFSET_BITS
#define _FILE_OFFSET_BITS	64
#endif

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdbool.h>
#include <stdatomic.h>
#include <inttypes.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <sys/mman.h>
#include <search.h>
#include <time.h>
#include <linux/fs.h>
#ifdef _OPENMP
#include <omp.h>
#endif

#include "endian.h"
#include "mounted.h"

#include "../struct.h"

/* fsck.c */

extern int ro;
extern int yes;
extern int preen;
extern int force;
extern int all;
extern int optimize_directories;
extern int optimized_some_dir;
extern const char *dev;
extern int h;
extern size_t size;
extern void *ptr;
#define sb	((struct nvfs_sb *)(ptr))
extern int uncorrected_errors;
extern int modified;

#define distance(x)	((uintptr_t)(x) - (uintptr_t)ptr)

/* fsckblockmap.c */

bool is_block_reserved(uint64_t block);
bool is_inode_reserved(uint64_t inode);

void *group_ptr(uint64_t group);

#define B_READ		0x01
#define B_SET		0x02
#define B_CLEAR		0x04
#define B_BLOCK		0x08
#define B_INODE		0x10
#define B_MEM		0x20
#define B_DEV		0x40

bool bmp_op(uint64_t b, unsigned flags);

bool inc_n_links(uint64_t ino);
void dec_n_links(uint64_t ino);
uint32_t get_n_links(uint64_t ino);

bool test_ino_flag(uint64_t ino);
void set_ino_flag(uint64_t ino);
void clear_ino_flag(uint64_t ino);

uint64_t alloc_block_or_inode(unsigned flags);

int init_blockmap(void);
void done_blockmap(void);

/* fsckblockvfy.c */

extern int verify_block_clear;
bool verify_block_range(uint64_t b);
bool verify_block(uint64_t ino, __le64 *p, bool clone);
void free_block(__le64 *p);

/* fsckcon.c */

int check_connectivity(void);

/* fsckdir.c */

uint64_t alloc_dir_leaf(uint64_t ino);
bool validate_directory(uint64_t ino, struct nvfs_inode *inode);

/* fsckdiradd.c */

struct nvfs_dir_entry *nvfs_find_de(uint64_t dir_ino, const char *name, unsigned namelen);
struct nvfs_dir_entry *add_directory_entry(uint64_t dir_ino, const char *name, unsigned namelen, uint64_t target_ino, unsigned char dtype);

/* fsckgroup.c */

int finalize_groups(void);

/* fscklog.c */

#define INODE_PROGRESS_STEP	65536
#define DIR_PROGRESS_STEP	65536
#define GROUP_PROGRESS_STEP	128
#define BLOCK_PROGRESS_STEP	1048576

#define MAX_AUTO_MESSAGES	256

#define Q_JUST_PRINT	0
#define Q_PREEN		1
#define Q_NORMAL	2
#define Q_NOAUTO	3

void fsck_log(const char *msg, ...) __attribute__((format(printf, 1, 2)));
void fsck_progress(unsigned pass, uint64_t total, uint64_t partial);
void fsck_internal_error(const char *msg, ...) __attribute__((format(printf, 1, 2), __noreturn__));

struct qkey {
	unsigned same;
	_Atomic unsigned auto_response;
	_Atomic uint64_t auto_count;
};

bool fsck_query(struct qkey *qkey, int level, const char *msg, const char *question, ...) __attribute__((format(printf, 3, 5)));

/* fsckinodes.c */

extern uint64_t n_dirs;

struct nvfs_inode *p_inode(uint64_t ino);
struct nvfs_group *p_group(uint64_t group);
int check_inodes(void);
int finalize_inodes(void);

/* fscklost.c */

int check_lost_blocks(void);

/* fsckmem.c */

void *fsck_malloc(size_t s);
void *fsck_realloc(void *p, size_t s);
void fsck_free(void *p);

void init_array(uint64_t **a, size_t *len);
void append_array(uint64_t **a, size_t *len, uint64_t val);

void *fsck_mmap(size_t size);
void fsck_munmap(void *p, size_t size);

/* fscksuper.c */

extern uint64_t n_groups;
extern uint64_t inodes_in_last_group;
extern uint64_t inode_bitmap_inodes;

extern unsigned log2_pages_per_group;
extern unsigned log2_inodes_per_group;
extern unsigned log2_bytes_per_group;

extern uint64_t max_bytes;

void superblock_modified(void);

int check_super(void);
